//******************************************************************************
// Designer     : Akashi Satoh
// Emp No.      : 
// E-Mail       : 
// Dept         : 
//
// Version      : 0.1
// Created      : Sep/2/2009
// Last Updated : Sep/15/2009
//
// File Name    : AES_MULTI.v
//******************************************************************************

`timescale 1ns/1ns


`define EBC_MODE 3'h0
`define CBC_MODE 3'h1
`define CFB_MODE 3'h2
`define OFB_MODE 3'h3
`define CTR_MODE 3'h4


`define IDLE     2'h0
`define KEYGEN   2'h1
`define CIPHER   2'h2

// ================================================================================
//       AES top module
// ================================================================================

module AES_MULTI( RSTn, CLK,
                  Kin, Din, Iin, Cin, Krdy, Irdy, Crdy, Drdy, EncDec, Kwidth, AESmode, EN,
                  Dout, Busy, Dvld, Kvld );

  // ================================================================================
  // Input/Output signals
  // ================================================================================
  input          RSTn;
  input          CLK;
  input  [127:0] Kin;               // Key input
  input  [127:0] Din;               // Data input
  input  [127:0] Iin;               // Initialization vector input
  input  [127:0] Cin;               // Counter input
  input          Krdy;              // Key input ready
  input          Irdy;              // Initialization vector input ready
  input          Crdy;              // Init Counter input ready
  input          Drdy;              // Data input ready
  input          EncDec;            // 0:Encryption 1:Decryption
  input    [1:0] Kwidth;            // Key select 00:128bit 10:192bit 11:256bit 
  input    [2:0] AESmode;           // 0:ECB  1:CBC  2:OFB  3:CFB  4:CTR
  input          EN;                // AES enable
  output [127:0] Dout;              // Data output
  output         Busy;              // AES busy
  output         Dvld;              // Data valid
  output         Kvld;              // Key valid

  // ================================================================================
  // Internal signals
  // ================================================================================
  // data buffers
  reg    [127:0] din_reg;           // Input DATA latch register
  reg    [127:0] r0;                // Round0 Key
  reg    [127:0] data_reg;          // Round1-14 DATA register
  reg    [127:0] data_out_reg;      // DATA output node
  reg    [127:0] init_vector_reg0;  // Initialization vector register0 for the CBC and CFB, OFB mode
  reg    [127:0] init_vector_reg1;  // Initialization vector register1 for the CBC and CFB, OFB mode
  reg    [127:0] counter_reg;
  reg    [127:0] keyEnc_reg0;       // Encryption KEY (  0-127bit) for the round 0
  reg    [127:0] keyEnc_reg1;       // Encryption KEY (128-255bit) for the round 0
  reg    [127:0] keyDec_reg0;       // Decryption KEY (  0-127bit) for the round 0
  reg    [127:0] keyDec_reg1;       // Decryption KEY (128-255bit) for the round 0
  reg    [127:0] keyTmp_reg0;       // Temporary KEY (  0-127bit) for the round 1-14
  reg    [127:0] keyTmp_reg1;       // Temporary KEY (128-255bit) for the round 1-14

// control
  wire           inv;               // 0:Encryption 1:Decryption
  reg            iv_top;            // 1:XOR IV with input data
  reg      [7:0] rc_reg;            // RC
  wire     [7:0] rc_next;
  reg     [14:0] round_reg;         // Round counter
  reg            data_valid_reg;
  reg            data_valid_tmp; 
  reg            key_valid_reg;

  reg    [127:0] data_out_node;
  wire   [127:0] data_next;
  wire   [127:0] key_init0;         // Initial Round KEY (  0-127bit) for the round 1-14
  wire   [127:0] key_init1;         // Initial Round KEY (128-255bit) for the round 1-14
  wire   [127:0] enc_key_next0, enc_key_next1;
  wire   [127:0] dec_key_next0, dec_key_next1;
  wire   [127:0] key_next0;
  wire   [127:0] key_next1;
  wire   [127:0] enc_round_key, dec_round_key;
  wire   [127:0] enc_data_next, dec_data_next;
  wire     [7:0] enc_rc_next, dec_rc_next;
  wire     [7:0] rc_final;
  reg            keyin_sel;         // 0:key0-127 1:key128-255
  reg            final_round;       // final round flag for enc and dec
  wire           final_roundX;      // final round-1 flag for enc and dec
  wire           final_roundK;      // final round flag for key generateion
  reg      [1:0] state_reg;
  reg            out_latch;

  reg    [127:0] round_data;
  reg            k128, k192, k256;

// =================================================================================================
// Equetions
// =================================================================================================
  // Key width decoder
  always @( Kwidth ) begin
    case ( Kwidth )
      2'h0 : { k128, k192, k256 } = 3'b100;  // 128bit-Key
      2'h1 : { k128, k192, k256 } = 3'b010;  // 192bit-Key
      2'h2 : { k128, k192, k256 } = 3'b001;  // 256bit-Key
      2'h3 : { k128, k192, k256 } = 3'b001;  // 256bit-Key
    endcase
  end

  // Encryption Unit
  EncUnit EncU ( .data( data_reg ), .key_add( enc_round_key ),
                 .final_round( final_round ), .data_next( enc_data_next )
               );

  // Decryption Unit
  DecUnit DecU ( .data( data_reg ), .key_add( dec_round_key ),
                 /*.count( round_reg ),*/ .round1( round_reg[1] ), .data_next( dec_data_next )
               );

  // Encryption Key genarate Unit
  EncKeyGen EncKeyGen( .mode( Kwidth ), .k128( k128 ), .k192( k192 ), .k256( k256 ),
                       .key0( keyTmp_reg0 ), .key1( keyTmp_reg1 ), .key_init0( key_init0 ), .key_init1( key_init1 ),
                       .count( round_reg ), .rc( rc_reg ), 
                       .key_add( enc_round_key ), .key_next0( enc_key_next0 ), .key_next1( enc_key_next1 ), .rc_next( enc_rc_next )
                     );

  // Decryption Key genarate Unit
  DecKeyGen DecKeyGen( .mode( Kwidth ), .k128( k128 ), .k192( k192 ), .k256( k256 ), .inv( inv ),
                       .key0( keyTmp_reg0 ), .key1( keyTmp_reg1 ), .key_init0( key_init0 ), .key_init1( key_init1 ),
                       .count( round_reg ),  .rc( rc_reg ), 
                       .key_add( dec_round_key ), .key_next0( dec_key_next0 ), .key_next1( dec_key_next1 ), .rc_next( dec_rc_next )
                     );

  // Chiper/Inv chiper select
  assign inv = (( EncDec == 1'b1 ) && ( state_reg != `KEYGEN ) && ( AESmode == `EBC_MODE ))? 'b1 : 1'b0
             | (( EncDec == 1'b1 ) && ( state_reg != `KEYGEN ) && ( AESmode == `CBC_MODE ))? 'b1 : 1'b0;

  //
  assign key_next0 = ( inv == 1'b0 )? enc_key_next0 : dec_key_next0;
  assign key_next1 = ( inv == 1'b0 )? enc_key_next1 : dec_key_next1;
  assign data_next = ( inv == 1'b0 )? enc_data_next : dec_data_next;
  assign rc_next =   ( inv == 1'b0 )? enc_rc_next : dec_rc_next;
  assign key_init0 = ( inv == 1'b0 )? keyEnc_reg0 : keyDec_reg0;
  assign key_init1 = ( inv == 1'b0 )? keyEnc_reg1 : keyDec_reg1;


  // Input Block.
  // Round0 data select
  always @( EncDec or AESmode or iv_top or Din or init_vector_reg0 or counter_reg or din_reg or data_next or data_out_reg ) begin
    case ( AESmode )
      `CBC_MODE : begin
                    if ( EncDec == 0 )
		                if ( iv_top == 1 ) r0 = Din ^ init_vector_reg0;
                      else r0 = Din ^ data_next;
                    else r0 = Din;
                  end
      `CFB_MODE : begin
                    if ( EncDec == 0 )
		                if ( iv_top == 1 ) r0 = init_vector_reg0;
                      else r0 = data_out_reg;
                    else
		                if ( iv_top == 1 ) r0 = init_vector_reg0;
                      else r0 = din_reg;
                  end
      `OFB_MODE : begin
                    if ( EncDec == 0 )
		                if ( iv_top == 1 ) r0 = init_vector_reg0;
                      else r0 = data_next;
                    else
		                if ( iv_top == 1 ) r0 = init_vector_reg0;
                      else r0 = data_next;
                  end
      `CTR_MODE : r0 = counter_reg;
        default : r0 = Din;             // ECB mode
    endcase
  end


  // Round data select
  always @( state_reg or r0 or key_init0 or data_next ) begin
    if ( state_reg == `IDLE ) round_data = r0 ^ key_init0;
    else round_data = data_next;
  end

  // Output select
  always @( EncDec or AESmode or din_reg or data_next or init_vector_reg1 ) begin
    case ( AESmode )
      `CBC_MODE : begin
                    if ( EncDec == 1'b1 ) data_out_node = data_next ^ init_vector_reg1;
                    else data_out_node = data_next;
                  end
      `CFB_MODE : data_out_node = data_next ^ din_reg;
      `OFB_MODE : data_out_node = data_next ^ din_reg;
      `CTR_MODE : data_out_node = data_next ^ din_reg;
        default : data_out_node = data_next;                 // EBC mode
    endcase
  end


  assign Busy = ( state_reg == `IDLE )? 1'b0: 1'b1;
  assign rc_final = (Kwidth == 2'h0)? 8'h36:  // 128
                    (Kwidth == 2'h1)? 8'h80:  // 192
                                    8'h40;
  assign final_roundX = (Kwidth == 2'h0)? round_reg[9]:  // final round -1
                        (Kwidth == 2'h1)? round_reg[11]:
                                          round_reg[13];
  assign final_roundK = (Kwidth == 2'h0)? round_reg[9]:
                        (Kwidth == 2'h1)? round_reg[11]:
                                          round_reg[12];

  assign Dout   = data_out_reg;
  assign Dvld = data_valid_reg;
  assign Kvld  = key_valid_reg;


  // Control state machine
  always @( posedge CLK or negedge RSTn ) begin
    if ( RSTn == 0 ) begin
      state_reg      <= `IDLE;
      final_round    <= 0;
      round_reg      <= 15'b000000000000001;
      iv_top         <= 0;
      data_valid_reg <= 0;
      key_valid_reg  <= 0;
      data_valid_tmp <= 0;
      keyin_sel      <= 0;
      rc_reg         <= 8'h00;
    end
    else if ( EN == 1'b1 ) begin
      if (( state_reg == `IDLE ) && ( Drdy == 1'b1 )) begin
        din_reg <= Din;
      end

      if ((( state_reg == `IDLE ) && ( Drdy == 1 )) || ( state_reg == `CIPHER )) begin
        data_reg <= round_data;
        iv_top   <= 0;
      end

      if ((Drdy == 1) || (final_round == 0)) begin
          keyTmp_reg0 <= key_next0;
          keyTmp_reg1 <= key_next1;
      end

      case ( state_reg )
        `IDLE:   begin
                   data_valid_tmp <= 0;

                   if ( Krdy == 1'b1 ) begin    // Generate Decryption key
                     final_round   <= 1'b0;
                     round_reg     <= 15'b000000000000001;
                     rc_reg        <= 8'h01;
                     key_valid_reg <= 1'b0;
                     if ( keyin_sel == 1'b0 ) begin
                       keyEnc_reg0 <= Kin;
                       if ( Kwidth == 2'b00 ) state_reg <= `KEYGEN;  // 128bit-Key
                       else keyin_sel <= 1'b1;                       // 192bit-Key, 256bit-Key
                     end
                     else begin
                       keyEnc_reg1 <= Kin;
                       keyin_sel   <= 1'b0;
                       state_reg   <= `KEYGEN;
                     end
                   end
                   else if (Drdy == 1'b1) begin
                     state_reg     <= `CIPHER;
                     final_round   <= 1'b0;
                     rc_reg        <= rc_next;
                     round_reg     <= round_reg << 1;

                     if ( EncDec == 1) begin
                       init_vector_reg0 <= Iin;
                       init_vector_reg1 <= init_vector_reg0;
                     end
                   end
                   else if ( Irdy == 1'b1 ) begin
                     init_vector_reg0 <= Iin;
                     init_vector_reg1 <= init_vector_reg0;
                     round_reg        <= 16'b0000000000000001;
                     iv_top <= 1'b1;
                     if ( inv == 1'b0 ) rc_reg  <= 8'h01;
                     else rc_reg  <= rc_final;
                   end
                   else if ( Crdy == 1'b1 ) counter_reg <= Cin;
                   else begin
                     if ( inv == 1'b0 ) rc_reg  <= 8'h01;
                     else rc_reg  <= rc_final;

                     data_valid_reg <= data_valid_tmp;
                   end
/*
                   if (( AESmode != `EBC_MODE ) && ((Irdy == 1) || ((Drdy == 1) && (EncDec == 1)))) begin
                     init_vector_reg0 <= Iin;
                     init_vector_reg1 <= init_vector_reg0;
                   end*/
                 end

        `KEYGEN: begin
                   if ( final_roundK == 1'b1 ) begin  // Key expantion completed
                     key_valid_reg <= 1;              // Key setup compreated
                     keyDec_reg0   <= key_next0;
                     keyDec_reg1   <= key_next1; 
                     state_reg     <= `IDLE;
                     round_reg     <= 15'b0000000_00000001;
                     rc_reg  <= 8'h01;
                   end 
                   else begin
                     rc_reg     <= rc_next;           // load next rc for key expantion
                     round_reg  <= round_reg << 1;
                     key_valid_reg <= 0;
                   end
                 end

        `CIPHER: begin
                   if ( final_roundX == 1'b1 ) begin   // Chip or InvChip expantion completed
                     final_round    <= 1'b1;
                     data_valid_reg <= 0;              // data encryption / decryption compreated
                     data_valid_tmp <= 1;
                     state_reg      <= `IDLE;
                     round_reg      <= 15'b0000000_00000001;

                     if ( inv == 1'b0 ) rc_reg  <= 8'h01;
                     else rc_reg  <= rc_final;

                     counter_reg <= counter_reg + 1'b1;   // CTR mode
                   end
                   else begin
                     final_round <= 0;
                     rc_reg      <= rc_next;
                     round_reg   <= round_reg << 1;
                   end
                 end
        endcase

        out_latch <= final_roundX;
        if ( out_latch == 1'b1 ) data_out_reg <= data_out_node;
      end
   end
endmodule