//******************************************************************************
// Designer     :
// Emp No.      : 
// E-Mail       : 
// Dept         : 
//
// Version      : 0.1
// Created      : Sep/2/2009
// Last Updated : Sep/15/2009
//
// File Name    : ENC_TOP.v
//******************************************************************************

`timescale 1ns/1ns

`define IDLE     2'h0
`define KEYGEN   2'h1
`define CIPHER   2'h2

// ================================================================================
//       AES top module
// ================================================================================

module ENC_TOP( RSTn, CLK,
                DinKin, Drdy, Krdy, Kwidth, enable,
                Dout, Busy, Dvld, Kvld );

  // ================================================================================
  // Input/Output signals
  // ================================================================================
  input          RSTn;
  input          CLK;
  input  [127:0] DinKin;            // Data or Key input
  input          Drdy;              // Data input ready
  input          Krdy;              // Key input ready
  input    [1:0] Kwidth;            // Key select 00:128bit 10:192bit 11:256bit 
  input          enable;            // AES enable
  output [127:0] Dout;              // Data output
  output         Busy;              // AES busy
  output         Dvld;              // Data valid
  output         Kvld;              // Key valid

  // ================================================================================
  // Internal signals
  // ================================================================================
  reg    [127:0] data_reg;          // Round1-14 DATA register
  reg    [127:0] data_out_reg;      // DATA output node
  reg    [127:0] EncKey_reg0;       // Encryption KEY (  0-127bit) for the round 0
  reg    [127:0] EncKey_reg1;       // Encryption KEY (128-255bit) for the round 0
  reg    [127:0] keyTmp_reg0;       // Temporary KEY (  0-127bit) for the round 1-14
  reg    [127:0] keyTmp_reg1;       // Temporary KEY (128-255bit) for the round 1-14

  reg      [7:0] rc_reg;            // RC
  wire     [7:0] rc_next;
  reg     [14:0] round_reg;         // Round counter
  reg            key_valid_reg;
  reg            data_valid_tmp; 
  reg            data_valid_reg;

  wire   [127:0] key_next0;
  wire   [127:0] data_next;
  wire   [127:0] key_next1;
  wire   [127:0] round_key;
  reg            keyin_sel;         // 0:key0-127 1:key128-255
  reg            final_round;       // final round flag for enc and dec
  wire           final_roundC;      // final round-1 flag for enc and dec
  wire           final_roundK;      // final round flag for key generateion
  reg      [1:0] state_reg;

  reg            k128, k192, k256;

// =================================================================================================
// Equetions
// =================================================================================================
  // Key width decoder
  always @( Kwidth ) begin
    case ( Kwidth )
      2'h0 : { k128, k192, k256 } = 3'b100;  // 128bit-Key
      2'h1 : { k128, k192, k256 } = 3'b010;  // 192bit-Key
      2'h2 : { k128, k192, k256 } = 3'b001;  // 256bit-Key
      2'h3 : { k128, k192, k256 } = 3'b001;  // 256bit-Key
    endcase
  end

  // Encryption Unit
  EncUnit EncU ( .data( data_reg ), .key_add( round_key ),
                 .final_round( final_round ), .data_next( data_next )
               );

  // Encryption Key genarate Unit
  EncKeyGen EncKey( .k128( k128 ), .k192( k192 ), .k256( k256 ),
                    .key0( keyTmp_reg0 ), .key1( keyTmp_reg1 ), .key_init0( EncKey_reg0 ), .key_init1( EncKey_reg1 ),
                    .mode( Kwidth ), .count( round_reg ), .rc( rc_reg ), 
                    .key_add( round_key ), .key_next0( key_next0 ), .key_next1( key_next1 ), .rc_next( rc_next)
                  );

  assign final_roundK = k128 & round_reg[9]
                      | k192 & round_reg[11]
                      | k256 & round_reg[12];
  assign final_roundC = k128 & round_reg[9]
                      | k192 & round_reg[11]
                      | k256 & round_reg[13];

  // Output
  assign Busy = ( state_reg == `IDLE )? 1'b0: 1'b1;
  assign Dout = data_out_reg;
  assign Dvld = data_valid_reg;
  assign Kvld = key_valid_reg;

  // Control state machine
  always @( posedge CLK or negedge RSTn ) begin
    if ( RSTn == 0 ) begin
      state_reg      <= `IDLE;
      data_reg       <= 128'h00000000_00000000_00000000_00000000;
      keyTmp_reg0    <= 128'h00000000_00000000_00000000_00000000;
      keyTmp_reg1    <= 128'h00000000_00000000_00000000_00000000;
      final_round    <= 1'b0;
      round_reg      <= 16'b0000000000000001;
      data_out_reg   <= 128'h00000000_00000000_00000000_00000000;
      data_valid_reg <= 1'b0;
      key_valid_reg  <= 1'b0;
      data_valid_tmp <= 1'b0;
      keyin_sel      <= 1'b0;
      rc_reg         <= 8'h00;
    end
    else if ( enable == 1'b1 ) begin
      if ((Drdy == 1) || (final_round == 0)) begin
        keyTmp_reg0 <= key_next0;
        keyTmp_reg1 <= key_next1;
      end

      case ( state_reg )
        `IDLE: begin
                 if ( Krdy == 1'b1 ) begin    // Generate Decryption key
                   final_round   <= 0;
                   round_reg     <= 16'b0000000000000001;
                   rc_reg        <= 8'h01;
                   key_valid_reg <= 1'b0;

                   if (keyin_sel == 0) begin
                     EncKey_reg0 <= DinKin;
                     if ( Kwidth == 2'b00 ) state_reg <= `KEYGEN;  // 128bit-Key
                     else keyin_sel <= 1'b1;                       // 192bit-Key, 256bit-Key
                   end
                   else begin
                     EncKey_reg1 <= DinKin;
                     keyin_sel   <= 1'b0;
                     state_reg   <= `KEYGEN;
                   end
                 end
                 else if ( Drdy == 1'b1 ) begin
                   data_reg <= DinKin ^ EncKey_reg0;//round_data;
                   data_valid_reg <= 1'b0;
                   final_round   <= 1'b0;
                   rc_reg        <= rc_next;
                   round_reg  <= round_reg << 1;
                   state_reg     <= `CIPHER;
                 end
                 else begin
                   rc_reg  <= 8'h01;
                 end

                 data_valid_tmp <= 0;
                 data_valid_reg <= data_valid_tmp;
             end

        `KEYGEN : begin
                    if ( final_roundK == 1'b1 ) begin  // Key expantion completed
                      key_valid_reg <= 1;              // Key setup compreated
                      round_reg     <= 16'b00000000_00000001;
                      rc_reg  <= 8'h01;
                      state_reg     <= `IDLE;
                    end 
                    else begin
                      rc_reg     <= rc_next;           // load next rc for key expantion
                      round_reg  <= round_reg << 1;
                      key_valid_reg <= 1'b0;
                    end
                  end

        `CIPHER: begin
                   if ( final_roundC == 1'b1 ) begin   // Chip or InvChip expantion completed
                     final_round    <= 1'b1;
                     data_valid_reg <= 1'b0;           // data encryption / decryption compreated
                     data_valid_tmp <= 1'b1;
                     state_reg      <= `IDLE;
                     round_reg      <= 16'b00000000_00000001;
                     rc_reg         <= 8'h01;
                   end
                   else begin
                     final_round <= 1'b0;
                     rc_reg      <= rc_next;
                     round_reg  <= round_reg << 1;
                   end

                   data_reg <= data_next;
                 end
        endcase

        if ( data_valid_tmp == 1'b1 ) data_out_reg <= data_next;
      end
   end
	
endmodule